@@ -2,10 +2,69 @@
 Parallel IO using MPI
 """
 
-__all__ = ["MpiIO", "Decomposition"]
+__all__ = ["MpiIO"]
 
+from contextlib import contextmanager
+import tempfile
+import os
 import numpy
 
+# pylint: disable=C0103
+try:
+    import mpi4py
+
+    with_mpi = True
+except ImportError:
+    with_mpi = False
+
+
+from .decomposition import Decomposition
+
+
+def check_comm(comm):
+    "Raises error if comm is not valid"
+    if not hasattr(comm, "size"):
+        raise TypeError(
+            "comm variable needs to be a valid MPI communicator with size attribute."
+        )
+
+
+def _tempdir_MPI(comm=None):
+    """
+    Creates a temporary directory to be used during testing
+    """
+
+    from mpi4py import MPI
+
+    if comm is None:
+        comm = MPI.COMM_WORLD
+    if comm.rank == 0:
+        tmp = tempfile.TemporaryDirectory()
+        name = tmp.__enter__()
+    else:
+        name = ""
+    path = comm.bcast(name, root=0)
+
+    # test path exists for all
+    has_access = os.path.exists(path) and os.access(path, os.R_OK | os.W_OK)
+    all_access = comm.allreduce(has_access, op=MPI.LAND)
+    if not all_access:
+        raise ValueError(
+            "Some processes are unable to access the temporary directory. \n\
+                Set TMPDIR, TEMP or TMP environment variables with the temporary \n\
+                directory to be used across processes. "
+        )
+
+    yield path + "/"
+
+    # make sure file exists until everyone is done
+    comm.Barrier()
+    if comm.rank == 0:
+        tmp.__exit__(None, None, None)
+
+
+tempdir_MPI = contextmanager(_tempdir_MPI)
+
 
 class MpiIO:
     """
@@ -19,7 +78,12 @@ class MpiIO:
         Property for importing MPI wherever necessary
         """
         # pylint: disable=C0415
-        from mpi4py import MPI
+        try:
+            from mpi4py import MPI
+        except ImportError as err:
+            raise ImportError(
+                "MPI not available. Consider installing `lyncs_io[mpi]`."
+            ) from err
 
         return MPI
 
@@ -73,11 +137,11 @@ class MpiIO:
         _, subsizes, _ = self.decomposition.decompose(domain)
         local_array = numpy.empty(subsizes, dtype=dtype, order=order.upper())
 
-        self.handler.Read_all(local_array)
+        self.handler.Read_all(self._array_view(local_array))
 
         return local_array
 
-    def save(self, array):
+    def save(self, array, header=None, offset=None):
         """
         Writes the local array in a file in parallel
 
@@ -92,17 +156,19 @@ class MpiIO:
         else:
             raise NotImplementedError("Currently noy supporting FORTRAN ordering")
 
-        if self.rank == 0:
-            pos = self.handler.Get_position()
-        else:
-            pos = 0
+        if offset is None:
+            if header:
+                offset = len(header)
+            else:
+                offset = 0
 
-        pos = self.comm.bcast(pos, root=0)
+        if self.rank == 0 and header:
+            self.handler.Write(header)
 
-        self._set_view(array.shape, numpy.array(array).dtype, "C", pos, compose=True)
+        self._set_view(array.shape, array.dtype, "C", offset, compose=True)
 
         # collectively write the array to file
-        self.handler.Write_all(array)
+        self.handler.Write_all(self._array_view(array))
 
     def _file_open(self, mode=None):
         if mode is None:
@@ -154,7 +220,11 @@ class MpiIO:
 
         return switcher.get(mode)
 
-    def _dtype_to_mpi(self, np_type):
+    def _array_view(self, array):
+        "Array view for MPI functions"
+        return array.view(self._dtype_to_mpi(array.dtype, get_key=True))
+
+    def _dtype_to_mpi(self, np_type, get_key=False):
         """
         Convert Numpy data type to MPI type
 
@@ -169,13 +239,26 @@ class MpiIO:
             MPI data type corresponding to `np_type`.
         """
 
+        np_type = numpy.dtype(np_type)
+        if np_type.byteorder == ">":
+            np_type = np_type.newbyteorder("<")
+
         if hasattr(self.MPI, "_typedict"):
-            mpi_type = self.MPI._typedict[numpy.dtype(np_type).char]
+            types = self.MPI._typedict
         elif hasattr(self.MPI, "__TypeDict__"):
-            mpi_type = self.MPI.__TypeDict__[numpy.dtype(np_type).char]
+            types = self.MPI.__TypeDict__
         else:
-            raise ValueError("cannot convert type")
-        return mpi_type
+            raise RuntimeError("Types dict not found")
+
+        for key, val in types.items():
+            try:
+                if numpy.dtype(key) == np_type:
+                    return key if get_key else val
+            # for keys that are not understood
+            except TypeError:
+                continue
+
+        raise TypeError(f"{np_type} is not supported")
 
     class _FileWrapper:
         """
@@ -189,155 +272,3 @@ class MpiIO:
             if key == "write":
                 key = "Write"
             return self.handler.__getattribute__(key)
-
-
-class Decomposition:
-    """
-    Decompose data using Cartesian/Domain Decomposition
-    of arbitrary dimensions
-    """
-
-    # pylint: disable=C0103
-    @property
-    def MPI(self):
-        """
-        Property for importing MPI wherever necessary
-        """
-        # pylint: disable=C0415
-        from mpi4py import MPI
-
-        return MPI
-
-    def __init__(self, comm=None):
-        if (comm is None) or (not isinstance(comm, self.MPI.Comm)):
-            raise TypeError("Expected an MPI communicator")
-
-        self.comm = comm
-        self.size = self.comm.Get_size()
-        self.rank = self.comm.Get_rank()
-
-        if comm.topology is self.MPI.GRAPH or comm.topology is self.MPI.DIST_GRAPH:
-            raise TypeError(f"comm is of unsupported type: {type(comm)}")
-        if comm.topology is self.MPI.CART:
-            self.dims = self.MPI.Compute_dims(self.size, comm.Get_dim())
-            self.coords = self.comm.Get_coords(self.rank)
-        elif isinstance(comm, self.MPI.Intracomm):
-            self.dims = [self.size]
-            self.coords = [self.rank]
-
-    def decompose(self, domain):
-        """
-        Decompose data over a cartesian/normal communicator.
-        Data domains of higher order compared to communicators order
-        only decomposed on the slow moving indexes.
-
-        Parameters
-        ----------
-        domain : list
-            Contains the global size domain we are decomposing.
-
-        Returns:
-        --------
-        sizes : list
-            global size of the domain
-        sub_sizes : list
-            local size of the domain
-        starts : list
-            global starting position
-        """
-
-        sizes = list(domain)
-        sub_sizes = list(domain)
-        starts = [0] * len(domain)
-
-        # Iterating over the dimensions of the topology
-        # allows for decomposition of higher order data domains
-        for dim in range(len(self.dims)):
-            workers, proc_id = self.dims[dim], self.coords[dim]
-            if domain[dim] < workers:
-                raise ValueError(
-                    "Domain size ({}) must be larger than the amount of workers({})".format(
-                        domain[dim], workers
-                    )
-                )
-
-            low = _split_work(domain[dim], workers, proc_id)
-            high = _split_work(domain[dim], workers, proc_id + 1)
-
-            sub_sizes[dim] = high - low
-            starts[dim] = low
-
-        return sizes, sub_sizes, starts
-
-    def compose(self, domain):
-        """
-        Reconstruct global data domain and position of the
-        local array relatively to the global domain.
-
-        Parameters
-        ----------
-        domain : list
-            Contains the local size domain of the array.
-
-        Returns:
-        --------
-        sizes : list
-            global size of the domain
-        sub_sizes : list
-            local size of the domain
-        starts : list
-            global starting position
-        """
-        sizes = list(domain)
-        sub_sizes = list(domain)
-        starts = [0] * len(domain)
-
-        # Iterating over the dimensions of the topology
-        # allows for composition of higher order data domains
-        for dim in range(len(self.dims)):
-            rdims = [False] * len(self.dims)
-            rdims[dim] = True
-
-            # sub-communicator for collective communications over
-            # a single dimension in the topology
-            if self.comm.topology is self.MPI.CART:
-                subcomm = self.comm.Sub(remain_dims=rdims)
-            else:
-                subcomm = self.comm
-
-            sub_size = subcomm.allgather(sub_sizes[dim])
-            sizes[dim] = sum(sub_size)
-            starts[dim] = numpy.cumsum([0] + sub_size)[self.coords[dim]]
-
-        return sizes, sub_sizes, starts
-
-
-def _split_work(load, workers, proc_id):
-    """
-    Uniformly distributes load over the dimension.
-    Remaining load is assigned in reverse round robin manner
-
-    Parameters
-    ----------
-    load : int
-        load size to be assigned
-    workers : int
-        total processing elements work will be assigned to
-    proc_id : int
-        processing element for which the bound is calculated for
-
-    Returns:
-    --------
-    bound : int
-        low bound of the workload assigned to the worker
-    """
-    unifload = load // workers  # uniform distribution
-    rem = load - unifload * workers
-
-    # round-robin assignment of the remaining work
-    if proc_id <= rem:
-        bound = (unifload + 1) * proc_id
-    else:
-        bound = (unifload + 1) * rem + unifload * (proc_id - rem)
-
-    return bound
